# Standard Library
from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system


class rpcs3(Runner):
    human_name = _("RPCS3")
    description = _("PlayStation 3 emulator")
    platforms = [_("Sony PlayStation 3")]
    runnable_alone = True
    runner_executable = "rpcs3/rpcs3"
    flatpak_id = "net.rpcs3.RPCS3"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "default_path": "game_path",
            "label": _("Path to EBOOT.BIN"),
        }
    ]
    runner_options = [{"option": "nogui", "type": "bool", "label": _("No GUI"), "default": False}]

    # RPCS3 currently uses an AppImage, no need for the runtime.
    system_options_override = [{"option": "disable_runtime", "default": True}]

    def play(self):
        arguments = self.get_command()

        if self.runner_config.get("nogui"):
            arguments.append("--no-gui")

        eboot = self.game_config.get("main_file") or ""
        if not system.path_exists(eboot):
            raise MissingGameExecutableError(filename=eboot)
        arguments.append(eboot)
        return {"command": arguments}



"""Manipulate Wine registry files"""

import os
import re
from collections import OrderedDict
from datetime import datetime

from lutris.util import system
from lutris.util.log import logger
from lutris.util.wine.wine import WINE_DEFAULT_ARCH

(
    REG_NONE,
    REG_SZ,
    REG_EXPAND_SZ,
    REG_BINARY,
    REG_DWORD,
    REG_DWORD_BIG_ENDIAN,
    REG_LINK,
    REG_MULTI_SZ,
) = range(8)

DATA_TYPES = {
    '"': REG_SZ,
    'str:"': REG_SZ,
    'str(2):"': REG_EXPAND_SZ,
    'str(7):"': REG_MULTI_SZ,
    "hex": REG_BINARY,
    "dword": REG_DWORD,
}


class WindowsFileTime:
    """Utility class to deal with Windows FILETIME structures.

    See: https://msdn.microsoft.com/en-us/library/ms724284(v=vs.85).aspx
    """

    ticks_per_seconds = 10000000  # 1 tick every 100 nanoseconds
    epoch_delta = 11644473600  # 3600 * 24 * ((1970 - 1601) * 365 + 89)

    def __init__(self, timestamp=None):
        self.timestamp = timestamp

    def __repr__(self):
        return "<{}>: {}".format(self.__class__.__name__, self.timestamp)

    @classmethod
    def from_hex(cls, hexvalue):
        timestamp = int(hexvalue, 16)
        return WindowsFileTime(timestamp)

    def to_hex(self):
        return "{:x}".format(self.timestamp)

    @classmethod
    def from_unix_timestamp(cls, timestamp):
        timestamp = timestamp + cls.epoch_delta
        timestamp = int(timestamp * cls.ticks_per_seconds)
        return WindowsFileTime(timestamp)

    def to_unix_timestamp(self):
        if not self.timestamp:
            raise ValueError("No timestamp set")
        unix_ts = self.timestamp / self.ticks_per_seconds
        unix_ts = unix_ts - self.epoch_delta
        return unix_ts

    def to_date_time(self):
        return datetime.fromtimestamp(self.to_unix_timestamp())


class WineRegistry:
    version_header = "WINE REGISTRY Version "
    relative_to_header = ";; All keys relative to "

    def __init__(self, reg_filename=None):
        self.arch = WINE_DEFAULT_ARCH
        self.version = 2
        self.relative_to = "\\\\User\\\\S-1-5-21-0-0-0-1000"
        self.keys = OrderedDict()
        self.reg_filename = reg_filename
        if reg_filename:
            if not system.path_exists(reg_filename):
                logger.error("No registry file at %s", reg_filename)
            self.parse_reg_file(reg_filename)

    def __str__(self):
        return "Windows Registry @ %s" % self.reg_filename

    @property
    def prefix_path(self):
        """Return the Wine prefix path (where the .reg files are located)"""
        if self.reg_filename:
            return os.path.dirname(self.reg_filename)
        return None

    @staticmethod
    def get_raw_registry(reg_filename):
        """Return an array of the unprocessed contents of a registry file"""
        if not system.path_exists(reg_filename):
            return []
        with open(reg_filename, "r", encoding="utf-8") as reg_file:
            try:
                registry_content = reg_file.readlines()
            except Exception:  # pylint: disable=broad-except
                logger.exception("Failed to registry read %s", reg_filename)
                registry_content = []
        return registry_content

    def parse_reg_file(self, reg_filename):
        registry_lines = self.get_raw_registry(reg_filename)
        current_key = None
        add_next_to_value = False
        additional_values = []
        for line in registry_lines:
            line = line.rstrip("\n")

            if line.startswith("["):
                current_key = WineRegistryKey(key_def=line)
                self.keys[current_key.name] = current_key
            elif current_key:
                if add_next_to_value:
                    additional_values.append(line)
                elif not add_next_to_value:
                    if additional_values:
                        additional_values = "\n".join(additional_values)
                        current_key.add_to_last(additional_values)
                        additional_values = []
                    current_key.parse(line)
                add_next_to_value = line.endswith("\\")
            elif line.startswith(self.version_header):
                self.version = int(line[len(self.version_header) :])
            elif line.startswith(self.relative_to_header):
                self.relative_to = line[len(self.relative_to_header) :]
            elif line.startswith("#arch"):
                self.arch = line.split("=")[1]

    def render(self):
        content = "{}{}\n".format(self.version_header, self.version)
        content += "{}{}\n\n".format(self.relative_to_header, self.relative_to)
        content += "#arch={}\n".format(self.arch)
        for key in self.keys:
            content += "\n"
            content += self.keys[key].render()
        return content

    def save(self, path=None):
        """Write the registry to a file"""
        if not path:
            path = self.reg_filename
        if not path:
            raise OSError("No filename provided")
        prefix_path = os.path.dirname(path)
        if not os.path.isdir(prefix_path):
            raise OSError(
                "Invalid Wine prefix path %s, make sure to create the prefix before saving to a registry" % prefix_path
            )
        with open(path, "w", encoding="utf-8") as registry_file:
            registry_file.write(self.render())

    def query(self, path, subkey):
        key = self.keys.get(path)
        if key:
            return key.get_subkey(subkey)
        return

    def set_value(self, path, subkey, value):
        key = self.keys.get(path)
        if not key:
            key = WineRegistryKey(path=path)
            self.keys[key.name] = key
        key.set_subkey(subkey, value)

    def clear_key(self, path):
        """Removes all subkeys from a key"""
        key = self.keys.get(path)
        if not key:
            return
        key.subkeys.clear()

    def clear_subkeys(self, path, keys):
        """Remove some subkeys from a key"""
        key = self.keys.get(path)
        if not key:
            return
        for subkey in list(key.subkeys.keys()):
            if subkey not in keys:
                continue
            key.subkeys.pop(subkey)

    def get_unix_path(self, windows_path):
        windows_path = windows_path.replace("\\", "/")
        if not self.prefix_path:
            return
        drives_path = os.path.join(self.prefix_path, "dosdevices")
        if not system.path_exists(drives_path):
            return
        letter, relpath = windows_path.split(":", 1)
        relpath = relpath.strip("/")
        drive_link = os.path.join(drives_path, letter.lower() + ":")
        try:
            drive_path = os.readlink(drive_link)
        except FileNotFoundError:
            logger.error("Unable to read link for %s", drive_link)
            return

        if not os.path.isabs(drive_path):
            drive_path = os.path.join(drives_path, drive_path)
        return os.path.join(drive_path, relpath)


class WineRegistryKey:
    def __init__(self, key_def=None, path=None):
        self.subkeys = OrderedDict()
        self.metas = OrderedDict()

        if path:
            # Key is created by path, it's a new key
            timestamp = datetime.now().timestamp()
            self.name = path
            self.raw_name = "[{}]".format(path.replace("/", "\\\\"))
            self.raw_timestamp = " ".join(str(timestamp).split("."))

            windows_timestamp = WindowsFileTime.from_unix_timestamp(timestamp)
            self.metas["time"] = windows_timestamp.to_hex()
        else:
            # Existing key loaded from file
            self.raw_name, self.raw_timestamp = re.split(re.compile(r"(?<=[^\\]\]) "), key_def, maxsplit=1)
            self.name = self.raw_name.replace("\\\\", "/").strip("[]")

        # Parse timestamp either as int or float
        ts_parts = self.raw_timestamp.strip().split()
        if len(ts_parts) == 1:
            self.timestamp = int(ts_parts[0])
        else:
            self.timestamp = float("{}.{}".format(ts_parts[0], ts_parts[1]))

    def __str__(self):
        return "{0} {1}".format(self.raw_name, self.raw_timestamp)

    def parse(self, line):
        """Parse a registry line, populating meta and subkeys"""
        if len(line) < 4:
            # Line is too short, nothing to parse
            return

        if line.startswith("#"):
            self.add_meta(line)
        elif line.startswith('"'):
            try:
                key, value = re.split(re.compile(r"(?<![^\\]\\\")="), line, maxsplit=1)
            except ValueError as ex:
                logger.error("Unable to parse line %s", line)
                logger.exception(ex)
                return
            key = key[1:-1]
            self.subkeys[key] = value
        elif line.startswith("@"):
            key, value = line.split("=", 1)
            self.subkeys["default"] = value

    def add_to_last(self, line):
        try:
            last_subkey = next(reversed(self.subkeys))
        except StopIteration:
            logger.warning("Should this be happening?")
            return
        self.subkeys[last_subkey] += "\n{}".format(line)

    def render(self):
        """Return the content of the key in the wine .reg format"""
        content = self.raw_name + " " + self.raw_timestamp + "\n"
        for key, value in self.metas.items():
            if value is None:
                content += "#{}\n".format(key)
            else:
                content += "#{}={}\n".format(key, value)
        for key, value in self.subkeys.items():
            if key == "default":
                key = "@"
            else:
                key = '"{}"'.format(key)
            content += "{}={}\n".format(key, value)
        return content

    def render_value(self, value):
        if isinstance(value, int):
            return "dword:{:08x}".format(value)
        if isinstance(value, str):
            return '"{}"'.format(value)
        raise NotImplementedError("TODO")

    @staticmethod
    def decode_unicode(string):
        # There may be a r"\\" in front of r"\x", so replace the r"\\" to r"\x005c"
        # to avoid missing matches. Example: r"C:\\users\\x1234\\\x0041\x0042CD".
        # Note the difference between r"\\x1234", r"\\\x0041" and r"\x0042".
        # It should be r"C:\users\x1234\ABCD" after decoding.
        chunks = re.split(r"\\x", string.replace(r"\\", r"\x005c"))
        out = chunks.pop(0).encode().decode("unicode_escape")
        for chunk in chunks:
            # We have seen file with unicode characters escaped on 1 byte (\xfa),
            # 1.5 bytes (\x444) and 2 bytes (\x00ed). So we try 0 padding, 1 and 2
            # (python wants its escaped sequence to be exactly on 4 characters).
            # The exception let us know if it worked or not
            for i in [0, 1, 2]:
                try:
                    out += "\\u{}{}".format("0" * i, chunk).encode().decode("unicode_escape")
                    break
                except UnicodeDecodeError:
                    pass
        return out

    def add_meta(self, meta_line):
        if not meta_line.startswith("#"):
            raise ValueError("Key metas should start with '#'")
        meta_line = meta_line[1:]
        parts = meta_line.split("=")
        if len(parts) == 2:
            key = parts[0]
            value = parts[1]
        elif len(parts) == 1:
            key = parts[0]
            value = None
        else:
            raise ValueError("Invalid meta line '{}'".format(meta_line))
        self.metas[key] = value

    def get_meta(self, name):
        return self.metas.get(name)

    def set_subkey(self, name, value):
        self.subkeys[name] = self.render_value(value)

    def get_subkey(self, name):
        if name not in self.subkeys:
            return None
        value = self.subkeys[name]
        if value.startswith('"') and value.endswith('"'):
            return self.decode_unicode(value[1:-1])
        if value.startswith("dword:"):
            return int(value[6:], 16)
        raise ValueError("Handle %s" % value)



"""Wine commands for installers"""

# pylint: disable=too-many-arguments
import os
import shlex
import time
from typing import Dict, Optional, Tuple

from lutris import runtime, settings
from lutris.config import LutrisConfig
from lutris.exceptions import MissingExecutableError
from lutris.monitored_command import MonitoredCommand
from lutris.runners import import_runner
from lutris.util import linux, system
from lutris.util.log import logger
from lutris.util.shell import get_shell_command
from lutris.util.strings import split_arguments
from lutris.util.wine import proton
from lutris.util.wine.cabinstall import CabInstaller
from lutris.util.wine.prefix import WinePrefixManager
from lutris.util.wine.wine import (
    WINE_DEFAULT_ARCH,
    WINE_DIR,
    detect_arch,
    get_overrides_env,
    get_real_executable,
    is_installed_systemwide,
    is_prefix_directory,
)


def set_regedit(
    path,
    key,
    value="",
    type="REG_SZ",  # pylint: disable=redefined-builtin
    wine_path=None,
    prefix=None,
    arch=WINE_DEFAULT_ARCH,
):
    """Add keys to the windows registry.

    Path is something like HKEY_CURRENT_USER/Software/Wine/Direct3D
    """
    formatted_value = {
        "REG_SZ": '"%s"' % value,
        "REG_DWORD": "dword:" + value,
        "REG_BINARY": "hex:" + value.replace(" ", ","),
        "REG_MULTI_SZ": "hex(2):" + value,
        "REG_EXPAND_SZ": "hex(7):" + value,
    }
    # Make temporary reg file
    reg_path = os.path.join(settings.CACHE_DIR, "winekeys.reg")
    with open(reg_path, "w", encoding="utf-8") as reg_file:
        reg_file.write('REGEDIT4\n\n[%s]\n"%s"=%s\n' % (path, key, formatted_value[type]))
    logger.debug("Setting [%s]:%s=%s", path, key, formatted_value[type])
    set_regedit_file(reg_path, wine_path=wine_path, prefix=prefix, arch=arch)
    os.remove(reg_path)


def set_regedit_file(filename, wine_path=None, prefix=None, arch=WINE_DEFAULT_ARCH, proton_verb=None):
    """Apply a regedit file to the Windows registry."""
    if arch == "win64" and wine_path and system.path_exists(wine_path + "64"):
        # Use wine64 by default if set to a 64bit prefix. Using regular wine
        # will prevent some registry keys from being created. Most likely to be
        # a bug in Wine. see: https://github.com/lutris/lutris/issues/804
        wine_path = wine_path + "64"

    if not wine_path or proton.is_proton_path(wine_path):
        proton_verb = "run"

    wineexec(
        "regedit",
        args="/S '%s'" % filename,
        wine_path=wine_path,
        prefix=prefix,
        arch=arch,
        blocking=True,
        proton_verb=proton_verb,
    )


def delete_registry_key(key, wine_path=None, prefix=None, arch=WINE_DEFAULT_ARCH, proton_verb=None):
    """Deletes a registry key from a Wine prefix"""

    if not wine_path or proton.is_proton_path(wine_path):
        proton_verb = "run"

    wineexec(
        "regedit",
        args='/S /D "%s"' % key,
        wine_path=wine_path,
        prefix=prefix,
        arch=arch,
        blocking=True,
        proton_verb=proton_verb,
    )


def create_prefix(
    prefix, wine_path=None, arch=WINE_DEFAULT_ARCH, overrides=None, install_gecko=None, install_mono=None, runner=None
):
    """Create a new Wine prefix."""
    if overrides is None:
        overrides = {}
    if not prefix:
        raise ValueError("No Wine prefix path given")
    prefix = os.path.expanduser(prefix)
    logger.info("Creating a %s prefix in %s", arch, prefix)

    # Follow symlinks, don't delete existing ones as it would break some setups
    if os.path.islink(prefix):
        prefix = os.readlink(prefix)

    if not runner:
        runner = import_runner("wine")(prefix=prefix, wine_arch=arch)

    if not wine_path:
        wine_path = runner.get_executable()

    logger.info("Winepath: %s", wine_path)

    if arch == "win32" and (proton.is_umu_path(wine_path) or proton.is_proton_path(wine_path)):
        logger.warning("Proton is not compatible with 32-bit prefixes, forcing win64")
        arch = "win64"

    wineenv = runner.system_config.get("env") or {}
    wineenv.update(
        {
            "WINEARCH": arch,
            "WINEPREFIX": prefix,
            "WINEDLLOVERRIDES": get_overrides_env(overrides),
            "WINE_MONO_CACHE_DIR": os.path.join(os.path.dirname(os.path.dirname(wine_path)), "mono"),
            "WINE_GECKO_CACHE_DIR": os.path.join(os.path.dirname(os.path.dirname(wine_path)), "gecko"),
        }
    )

    if install_gecko == "False":
        wineenv["WINE_SKIP_GECKO_INSTALLATION"] = "1"
        overrides["mshtml"] = "disabled"
    if install_mono == "False":
        wineenv["WINE_SKIP_MONO_INSTALLATION"] = "1"
        overrides["mscoree"] = "disabled"

    if proton.is_umu_path(wine_path) or proton.is_proton_path(wine_path):
        # All proton path prefixes are created via Umu; if you aren't using
        # the default Umu, we'll use PROTONPATH to indicate what Proton is
        # to be used.
        wineenv["PROTON_VERB"] = "run"

        proton.update_proton_env(wine_path, wineenv)

        umu_command = MonitoredCommand([proton.get_umu_path(), "createprefix"], env=wineenv)
        umu_command.start()
    else:
        umu_command = None
        wineboot_path = os.path.join(os.path.dirname(wine_path), "wineboot")
        if not system.path_exists(wineboot_path):
            logger.error(
                "No wineboot executable found in %s, your wine installation is most likely broken",
                wine_path,
            )
            return

        system.execute([wineboot_path], env=wineenv)

    for loop_index in range(1000):
        time.sleep(0.5)
        if (
            system.path_exists(os.path.join(prefix, "user.reg"))
            and system.path_exists(os.path.join(prefix, "userdef.reg"))
            and system.path_exists(os.path.join(prefix, "system.reg"))
        ):
            break
        # Check if umu crashed before prefix was created
        if umu_command and not umu_command.is_running:
            logger.error("Umu exited unexpectedly during prefix creation (return code: %s)", umu_command.return_code)
            return
        if loop_index == 60:
            logger.warning("Wine prefix creation is taking longer than expected...")
    if not os.path.exists(os.path.join(prefix, "user.reg")):
        logger.error("No user.reg found after prefix creation. Prefix might not be valid")
        return
    logger.info("%s Prefix created in %s", arch, prefix)
    prefix_manager = WinePrefixManager(prefix)
    prefix_manager.setup_defaults()


def winekill(prefix, arch=WINE_DEFAULT_ARCH, wine_path="", env=None, initial_pids=None, runner=None):
    """Kill processes in Wine prefix."""

    initial_pids = initial_pids or []
    if not env:
        env = {
            "WINEARCH": arch,
            "WINEPREFIX": prefix,
            "GAMEID": proton.DEFAULT_GAMEID,
        }
    env["PROTON_VERB"] = "runinprefix"  # must not block until the game exits, that would be sily!
    if proton.is_umu_path(wine_path):
        command = [wine_path, "wineboot", "-k"]
    elif proton.is_proton_path(wine_path):
        command = [proton.get_umu_path(), "wineboot", "-k"]
        env["PROTONPATH"] = proton.get_proton_path_by_path(wine_path)
    else:
        if not wine_path:
            if not runner:
                runner = import_runner("wine")()
            wine_path = runner.get_executable()
        wine_root = os.path.dirname(wine_path)

        command = [os.path.join(wine_root, "wineserver"), "-k"]
        logger.debug("Killing all wine processes (%s) in prefix %s: %s", initial_pids, prefix, command)
    logger.debug(command)
    logger.debug(" ".join(command))
    system.execute(command, env=env, quiet=True)

    logger.debug("Waiting for wine processes to terminate")
    # Wineserver needs time to terminate processes
    num_cycles = 0
    while True:
        num_cycles += 1
        running_processes = [pid for pid in initial_pids if system.path_exists("/proc/%s" % pid)]

        if not running_processes:
            break
        if num_cycles > 20:
            logger.warning("Some wine processes are still running: %s", running_processes)
            break
        time.sleep(0.1)
    logger.debug("Done waiting.")


def use_lutris_runtime(wine_path, force_disable=False):
    """Returns whether to use the Lutris runtime.
    The runtime can be forced to be disabled, otherwise
    it's disabled automatically if Wine is installed system wide.
    """
    if proton.is_proton_path(wine_path):
        return False
    if force_disable or runtime.RUNTIME_DISABLED:
        return False
    if WINE_DIR in wine_path:
        return True
    if is_installed_systemwide():
        return False
    return True


def wineexec(
    executable: str,
    prefix: str,
    args: str = "",
    wine_path: Optional[str] = None,
    arch: str = WINE_DEFAULT_ARCH,
    working_dir: Optional[str] = None,
    winetricks_wine: str = "",
    blocking: bool = False,
    config: Optional[LutrisConfig] = None,
    include_processes: Optional[list] = None,
    exclude_processes: Optional[list] = None,
    disable_runtime: bool = False,
    env: Optional[dict] = None,
    overrides=None,
    runner=None,
    proton_verb: Optional[str] = None,
):
    """
    Execute a Wine command.

    Args:
        executable (str): wine program to run, pass None to run wine itself
        prefix (str): path to the wine prefix to use
        args (str): program arguments
        wine_path (str): path to the wine version to use
        arch (str): wine architecture of the prefix
        working_dir (str): path to the working dir for the process
        winetricks_wine (str): path to the wine version used by winetricks
        blocking (bool): if true, do not run the process in a thread
        config (LutrisConfig): LutrisConfig object for the process context
        watch (list): list of process names to monitor (even when in a ignore list)
        runner (runner): the wine runner that carries the configuration to use

    Returns:
        Process results if the process is running in blocking mode or
        MonitoredCommand instance otherwise.
    """
    env = env or {}
    exclude_processes = exclude_processes or []
    include_processes = include_processes or []
    executable = str(executable) if executable else ""
    if isinstance(include_processes, str):
        include_processes = shlex.split(include_processes)
    if isinstance(exclude_processes, str):
        exclude_processes = shlex.split(exclude_processes)

    if not runner:
        runner = import_runner("wine")(prefix=prefix, working_dir=working_dir, wine_arch=arch)

    if not wine_path:
        wine_path = runner.get_executable()

        if not wine_path:  # to satisfy mypy really
            raise MissingExecutableError("The wine path could not be determined.")

    if arch == "win32" and (proton.is_umu_path(wine_path) or proton.is_proton_path(wine_path)):
        logger.warning("Proton is not compatible with 32-bit prefixes, forcing win64")
        arch = "win64"

    if not working_dir:
        if os.path.isfile(executable):
            working_dir = os.path.dirname(executable)

    executable, _args, working_dir = get_real_executable(executable, working_dir)
    if _args:
        args = '{} "{}"'.format(_args[0], _args[1])

    wineenv = {"WINEARCH": arch}
    if winetricks_wine and winetricks_wine is not wine_path and not proton.is_proton_path(wine_path):
        wineenv["WINE"] = winetricks_wine
    elif wine_path:
        wineenv["WINE"] = wine_path

    if prefix:
        wineenv["WINEPREFIX"] = prefix

    # Create prefix if necessary
    if arch not in ("win32", "win64"):
        arch = detect_arch(prefix, wine_path)
    if not is_prefix_directory(prefix):
        wine_bin = winetricks_wine if winetricks_wine and not proton.is_proton_path(wine_path) else wine_path
        create_prefix(prefix, wine_path=wine_bin, arch=arch, runner=runner)

    wine_system_config = config.system_config if config else runner.system_config
    disable_runtime = disable_runtime or wine_system_config["disable_runtime"]
    if use_lutris_runtime(wine_path=wineenv["WINE"], force_disable=disable_runtime) and not proton.is_proton_path(
        wine_path
    ):
        if WINE_DIR in wine_path:
            wine_root_path = os.path.dirname(os.path.dirname(wine_path))
        elif winetricks_wine and WINE_DIR in winetricks_wine:
            wine_root_path = os.path.dirname(os.path.dirname(winetricks_wine))
        else:
            wine_root_path = None
        wineenv["LD_LIBRARY_PATH"] = ":".join(
            runtime.get_paths(
                prefer_system_libs=wine_system_config["prefer_system_libs"],
                wine_path=wine_root_path,
            )
        )

    if overrides:
        wineenv["WINEDLLOVERRIDES"] = get_overrides_env(overrides)

    if proton_verb:
        wineenv["PROTON_VERB"] = proton_verb

    baseenv = runner.get_env(disable_runtime=disable_runtime)
    baseenv.update(wineenv)
    baseenv.update(env)

    if proton.is_proton_path(wine_path):
        proton.update_proton_env(wine_path, baseenv)

    command_parameters = []
    if proton.is_proton_path(wine_path):
        command_parameters.append(proton.get_umu_path())
        if winetricks_wine and wine_path not in winetricks_wine and "winetricks" not in args:
            command_parameters.append("winetricks")
    else:
        command_parameters.append(wine_path)
    if executable:
        command_parameters.append(executable)
    command_parameters += split_arguments(args)

    runner.prelaunch()

    if blocking:
        return system.execute(command_parameters, env=baseenv, cwd=working_dir)

    command = MonitoredCommand(
        command_parameters,
        runner=runner,
        env=baseenv,
        cwd=working_dir,
        include_processes=include_processes,
        exclude_processes=exclude_processes,
    )
    command.start()
    return command


# pragma pylint: enable=too-many-locals


def find_winetricks(
    env: Optional[Dict[str, str]] = None, system_winetricks: bool = False
) -> Tuple[str, Optional[str], Dict[str, str]]:
    """Find winetricks path."""
    env = env or {}
    winetricks_path = os.path.join(settings.RUNTIME_DIR, "winetricks/winetricks")
    if system_winetricks or not system.path_exists(winetricks_path):
        winetricks_path = system.find_required_executable("winetricks")
        working_dir = None
    else:
        # We will use our own zenity if available, which is here, and it
        # also needs a data file in this directory. We have to set the
        # working_dir, so it will find the data file.
        working_dir = os.path.join(settings.RUNTIME_DIR, "winetricks")

        path = env.get("PATH", os.environ["PATH"])
        env["PATH"] = "%s:%s" % (working_dir, path)

    return (winetricks_path, working_dir, env)


def winetricks(
    app: Optional[str],
    prefix: str,
    arch: str = WINE_DEFAULT_ARCH,
    silent: bool = True,
    wine_path: Optional[str] = None,
    config=None,
    env=None,
    disable_runtime=False,
    system_winetricks=False,
    runner=None,
    proton_verb=None,
):
    """Execute winetricks."""
    if not app:
        silent = False
        app = "--gui"
    args = app

    if wine_path and proton.is_proton_path(wine_path):
        protonfixes_path = os.path.join(proton.get_proton_path_by_path(wine_path), "protonfixes")
    else:
        protonfixes_path = None

    if protonfixes_path and os.path.exists(protonfixes_path):
        proton_verb = "waitforexitandrun"
        working_dir = None
        winetricks_wine = os.path.join(protonfixes_path, "winetricks")
        winetricks_path = wine_path
    elif not wine_path or proton.is_umu_path(wine_path):
        winetricks_wine = proton.get_umu_path()
        winetricks_path = None
        args = "winetricks " + args
        proton_verb = "waitforexitandrun"
        working_dir = None
    else:
        if protonfixes_path:
            logger.warning(
                "winetricks: attempting to run on a Valve official Proton build; this may not work as expected."
            )
        winetricks_path, working_dir, env = find_winetricks(env, system_winetricks)
        if not runner:
            runner = import_runner("wine")()
        winetricks_wine = runner.get_executable()
        if arch not in ("win32", "win64"):
            arch = detect_arch(prefix, winetricks_wine)
        if str(silent).lower() in ("yes", "on", "true"):
            args = "-q " + args

    # Execute wineexec
    return wineexec(
        "",
        prefix=prefix,
        winetricks_wine=winetricks_wine,
        wine_path=winetricks_path,
        working_dir=working_dir,
        arch=arch,
        args=args,
        config=config,
        env=env,
        disable_runtime=disable_runtime,
        runner=runner,
        proton_verb=proton_verb,
    )


def winecfg(wine_path=None, prefix=None, arch=WINE_DEFAULT_ARCH, config=None, env=None, runner=None, proton_verb=None):
    """Execute winecfg."""

    if not wine_path:
        logger.debug("winecfg: Reverting to default wine")
        wine = import_runner("wine")
        wine_path = wine().get_executable()

    if proton.is_proton_path(wine_path):
        proton_verb = "waitforexitandrun"

    return wineexec(
        "winecfg.exe",
        prefix=prefix,
        winetricks_wine=wine_path,
        wine_path=wine_path,
        arch=arch,
        config=config,
        env=env,
        include_processes=["winecfg.exe"],
        runner=runner,
        proton_verb=proton_verb,
    )


def eject_disc(wine_path: str, prefix: str, proton_verb=None):
    """Use Wine to eject a drive"""

    if proton.is_proton_path(wine_path):
        proton_verb = "run"
    wineexec("eject", prefix=prefix, wine_path=wine_path, args="-a", proton_verb=proton_verb)


def install_cab_component(cabfile, component, wine_path: str, prefix=None, arch=None, proton_verb=None):
    """Install a component from a cabfile in a prefix"""

    if proton.is_proton_path(wine_path):
        proton_verb = "run"
    cab_installer = CabInstaller(prefix, wine_path=wine_path, arch=arch)
    files = cab_installer.extract_from_cab(cabfile, component)
    registry_files = cab_installer.get_registry_files(files)
    for registry_file, _arch in registry_files:
        set_regedit_file(registry_file, wine_path=wine_path, prefix=prefix, arch=_arch, proton_verb=proton_verb)
    cab_installer.cleanup()


def open_wine_terminal(
    terminal: Optional[str], wine_path: str, prefix: str, env: Optional[Dict[str, str]], system_winetricks: bool
):
    winetricks_path, _working_dir, env = find_winetricks(env, system_winetricks)
    path_paths = [os.path.dirname(wine_path)]
    if proton.is_proton_path(wine_path):
        proton.update_proton_env(wine_path, env)
        umu_path = proton.get_umu_path()
        path_paths.insert(0, os.path.dirname(umu_path))
        wine_command = umu_path + " wine"
    else:
        wine_command = wine_path

    aliases = {
        "wine": wine_command,
        "winecfg": wine_command + "cfg",
        "wineserver": wine_command + "server",
        "wineboot": wine_command + "boot",
        "winetricks": winetricks_path,
    }
    env["WINEPREFIX"] = prefix
    # Ensure scripts you run see the desired version of WINE too
    # by putting it on the PATH.

    path_paths.append(env.get("PATH", os.environ["PATH"]))
    path_paths = [p for p in path_paths if p]
    if path_paths:
        env["PATH"] = ":".join(path_paths)
    shell_command = get_shell_command(prefix, env, aliases)
    terminal = terminal or linux.get_required_default_terminal()
    system.spawn([terminal, "-e", shell_command])



import glob
import os
import re
import shutil
import subprocess
from typing import Dict, Optional

from lutris.util import system
from lutris.util.graphics import drivers
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger

VULKANINFO_AVAILABLE = shutil.which("vulkaninfo")
VULKAN_DATA_DIRS = [
    "/usr/local/etc",  # standard site-local location
    "/usr/local/share",  # standard site-local location
    "/etc",  # standard location
    "/usr/share",  # standard location
    "/usr/lib/x86_64-linux-gnu/GL",  # Flatpak GL extension
    "/usr/lib/i386-linux-gnu/GL",  # Flatpak GL32 extension
    "/opt/amdgpu-pro/etc",  # AMD GPU Pro - TkG
]

GPUS = {}


def get_gpus_info():
    """Return the information related to each GPU on the system"""
    return {card: drivers.get_gpu_info(card) for card in drivers.get_gpu_cards()}


def display_gpu_info(gpu_id, gpu_info):
    """Log GPU information"""
    try:
        gpu_string = f"GPU: {gpu_info['PCI_ID']} {gpu_info['PCI_SUBSYS_ID']} ({gpu_info['DRIVER']} drivers)"
        logger.info(gpu_string)
    except KeyError:
        logger.error("Unable to get GPU information from '%s'", gpu_id)


def add_icd_search_path(paths):
    icd_paths = []
    if paths:
        # unixy env vars with multiple paths are : delimited
        for path in paths.split(":"):
            path = os.path.join(path, "vulkan")
            if os.path.exists(path) and path not in icd_paths:
                icd_paths.append(path)
    return icd_paths


def get_vk_icd_files():
    """Returns available vulkan ICD files in the same search order as vulkan-loader,
    but in a single list"""
    icd_search_paths = []
    for path in VULKAN_DATA_DIRS:
        icd_search_paths += add_icd_search_path(path)
    all_icd_files = []
    for data_dir in icd_search_paths:
        path = os.path.join(data_dir, "icd.d", "*.json")
        # sort here as directory enumeration order is not guaranteed in linux
        # so it's consistent every time
        icd_files = sorted(glob.glob(path))
        if icd_files:
            all_icd_files += icd_files
    return all_icd_files


class GPU:
    def __init__(self, card):
        self.card = card
        self.gpu_info = self.get_gpu_info()
        self.driver = self.gpu_info["DRIVER"]
        self.pci_id = self.gpu_info["PCI_ID"].lower()
        self.pci_subsys_id = self.gpu_info["PCI_SUBSYS_ID"].lower()
        self.pci_slot = self.gpu_info["PCI_SLOT_NAME"]
        self.icd_files = self.get_icd_files()
        if VULKANINFO_AVAILABLE:
            try:
                self.device_uuid = self.get_vulkaninfo_device_uuid()
                self.name = self.get_vulkaninfo_name() or self.get_lspci_name()
            except (OSError, subprocess.CalledProcessError, subprocess.TimeoutExpired):
                # already logged this, so we'll just fall back to lspci.
                self.name = self.get_lspci_name()
        else:
            self.name = self.get_lspci_name()

    def __str__(self):
        if self.pci_id:
            return f"{self.short_name} ({self.pci_id} {self.pci_subsys_id} {self.driver})"

        return f"{self.short_name} ({self.driver})"

    def get_driver_info(self):
        driver_info = {}
        if self.driver == "nvidia":
            driver_info = drivers.get_nvidia_driver_info()
        elif LINUX_SYSTEM.glxinfo:
            if hasattr(LINUX_SYSTEM.glxinfo, "GLX_MESA_query_renderer"):
                driver_info = {
                    "vendor": LINUX_SYSTEM.glxinfo.opengl_vendor,
                    "version": LINUX_SYSTEM.glxinfo.GLX_MESA_query_renderer.version,
                    "device": LINUX_SYSTEM.glxinfo.GLX_MESA_query_renderer.device,
                }
        return driver_info

    def get_gpu_info(self) -> Dict[str, str]:
        """Return information about a GPU"""
        infos = {"DRIVER": "", "PCI_ID": "", "PCI_SUBSYS_ID": "", "PCI_SLOT_NAME": ""}
        try:
            with open(f"/sys/class/drm/{self.card}/device/uevent", encoding="utf-8") as card_uevent:
                content = card_uevent.readlines()
        except FileNotFoundError:
            logger.error("Unable to read driver information for card %s", self.card)
            raise
        for line in content:
            key, value = line.split("=", 1)
            infos[key] = value.strip()
        return infos

    def get_vulkaninfo(self) -> Dict[str, Dict[str, str]]:
        """Runs vulkaninfo to find the GPU name"""
        subprocess_env = dict(os.environ)
        vulkaninfo_output_raw = system.read_process_output(
            ["/usr/bin/vulkaninfo", "--summary"], env=os.environ, error_result=None
        )
        if not vulkaninfo_output_raw:
            subprocess_env["VK_DRIVER_FILES"] = self.icd_files  # Currently supporte
            subprocess_env["VK_ICD_FILENAMES"] = self.icd_files  # Deprecated
            vulkaninfo_output_raw = system.read_process_output(
                ["/usr/bin/vulkaninfo", "--summary"], env=subprocess_env, error_result=""
            )

        vulkaninfo_output = vulkaninfo_output_raw.split("\n") if vulkaninfo_output_raw else []
        result = {}
        devices_seen = False
        for line in vulkaninfo_output:
            line = line.strip()
            if not line or line.startswith("==="):
                continue
            if line == "Devices:":
                devices_seen = True
                continue
            if not devices_seen:
                continue
            if line.startswith("GPU"):
                current_gpu = line
                result[current_gpu] = {}
            elif "= " in line:
                key, value = line.split("= ", maxsplit=1)
                result[current_gpu][key.strip()] = value.strip()
        if "Failed to detect any valid GPUs" in result or "ERROR: [Loader Message]" in result:
            logger.warning("Vulkan failed to detect any GPUs: %s", result)
            return {}
        return result

    def get_vulkaninfo_name(self) -> Optional[str]:
        vulkaninfo = self.get_vulkaninfo()
        best_name = None
        for gpu_index in vulkaninfo:
            pci_id = "%s:%s" % (
                vulkaninfo[gpu_index]["vendorID"].replace("0x", ""),
                vulkaninfo[gpu_index]["deviceID"].replace("0x", ""),
            )
            if pci_id == self.pci_id:
                name = vulkaninfo[gpu_index]["deviceName"]
                if not best_name or len(name) > len(best_name):
                    best_name = name
        return best_name

    def get_vulkaninfo_device_uuid(self) -> Optional[str]:
        vulkaninfo = self.get_vulkaninfo()
        for gpu_index in vulkaninfo:
            pci_id = "%s:%s" % (
                vulkaninfo[gpu_index]["vendorID"].replace("0x", ""),
                vulkaninfo[gpu_index]["deviceID"].replace("0x", ""),
            )
            if pci_id == self.pci_id:
                return vulkaninfo[gpu_index]["deviceUUID"].replace("-", "")
        return None

    def get_lspci_name(self):
        lspci_results = [line.split(maxsplit=1) for line in system.execute(["lspci"], timeout=3).split("\n")]
        lspci_results = [parts for parts in lspci_results if len(parts) == 2 and ": " in parts[1]]
        devices = [(pci_id, device_desc.split(": ")[1]) for pci_id, device_desc in lspci_results]
        for device in devices:
            if f"0000:{device[0]}" == self.pci_slot:
                return device[1]
        return "No GPU"

    def get_icd_files(self):
        loader = self.driver
        loader_map = {
            "amdgpu": "radeon",
            "vc4-drm": "broadcom",
            "v3d": "broadcom",
            "virtio-pci": "lvp",
            "i915": "intel",
            "xe": "intel",
        }
        if self.driver in loader_map:
            loader = loader_map[self.driver]
        icd_files = []
        for icd_file in get_vk_icd_files():
            if loader in icd_file:
                icd_files.append(icd_file)
        return ":".join(icd_files)

    @property
    def short_name(self):
        """Shorten result to just the friendly name of the GPU
        vulkaninfo returns Vendor Friendly Name (Chip Developer Name)
        AMD Radeon Pro W6800 (RADV NAVI21) -> AMD Radeon Pro W6800"""
        return re.sub(r"\s*\(.*?\)", "", self.name)



"""Install script interpreter package."""

import enum

import yaml

from lutris.api import get_game_installers, normalize_installer
from lutris.util import system
from lutris.util.log import logger

AUTO_EXE_PREFIX = "_xXx_AUTO_"
AUTO_ELF_EXE = AUTO_EXE_PREFIX + "ELF_xXx_"
AUTO_WIN32_EXE = AUTO_EXE_PREFIX + "WIN32_xXx_"


class InstallationKind(enum.Enum):
    INSTALL = 0
    UPDATE = 1
    DLC = 2


def read_script(filename):
    """Return scripts from a local file"""
    logger.debug("Loading script(s) from %s", filename)
    with open(filename, "r", encoding="utf-8") as local_file:
        script = yaml.safe_load(local_file.read())
        if isinstance(script, list):
            return script
        if "results" in script:
            return script["results"]
        return [script]


def get_installers(game_slug=None, installer_file=None, revision=None):
    # check if installer is local or online
    if system.path_exists(installer_file):
        return [normalize_installer(i, installer_file=installer_file) for i in read_script(installer_file)]
    return get_game_installers(game_slug=game_slug, revision=revision)



import os

from lutris.config import write_game_config
from lutris.database.games import add_game, get_games
from lutris.util.log import logger
from lutris.util.retroarch.core_config import RECOMMENDED_CORES
from lutris.util.strings import slugify

SCANNERS = {
    "mesen": "NES",
    "gambatte": "Gameboy / Gameboy Color",
    "snes": "Super Nintendo",
    "mupen64plus_next": "Nintendo 64",
    "picodrive": "Master System / Game Gear / Genesis / MegaCD / 32x",
    "opera": "3DO",
}


ROM_FLAGS = [
    "USA",
    "Europe",
    "World",
    "Japan",
    "Japan, USA",
    "USA, Europe",
    "Proto",
    "SGB Enhanced",
    "Rev A",
    "V1.1",
    "F",
    "U",
    "E",
    "UE",
    "W",
    "M3",
]

EXTRA_FLAGS = ["!", "S"]


def clean_rom_name(name):
    """Remove known flags from ROM filename and apply formatting"""
    for flag in ROM_FLAGS:
        name = name.replace(" (%s)" % flag, "")
    for flag in EXTRA_FLAGS:
        name = name.replace("[%s]" % flag, "")
    if ", The" in name:
        name = "The %s" % name.replace(", The", "")
    name = name.strip()
    return name


def scan_directory(dirname):
    """Add a directory of ROMs as Lutris games"""
    files = os.listdir(dirname)
    folder_extensions = {os.path.splitext(filename)[1] for filename in files}
    core_matches = {}
    for core, core_data in RECOMMENDED_CORES.items():
        for ext in core_data.get("extensions", []):
            if ext in folder_extensions:
                core_matches[ext] = core
    added_games = []
    for filename in files:
        name, ext = os.path.splitext(filename)
        if ext not in core_matches:
            continue
        logger.info("Importing '%s'", name)
        slug = slugify(name)
        core = core_matches[ext]
        config = {"game": {"core": core_matches[ext], "main_file": os.path.join(dirname, filename)}}
        installer_slug = "%s-libretro-%s" % (slug, core)
        existing_game = get_games(filters={"installer_slug": installer_slug})
        if existing_game:
            continue
        configpath = write_game_config(slug, config)
        game_id = add_game(
            name=name,
            runner="libretro",
            slug=slug,
            directory=dirname,
            installed=1,
            installer_slug=installer_slug,
            configpath=configpath,
        )
        added_games.append(game_id)
    return added_games



"""System utilities"""

import hashlib
import os
import re
import shutil
import stat
import string
import subprocess
import zipfile
from gettext import gettext as _
from pathlib import Path
from typing import IO, Dict, Iterable, List, Mapping, Optional, Sequence, Set, Tuple, Union

from gi.repository import Gio, GLib

from lutris import settings
from lutris.exceptions import MissingExecutableError
from lutris.util.log import logger
from lutris.util.portals import TrashPortal

# Home folders that should never get deleted.
PROTECTED_HOME_FOLDERS = (
    _("Documents"),
    _("Downloads"),
    _("Desktop"),
    _("Pictures"),
    _("Videos"),
    _("Pictures"),
    _("Projects"),
    _("Games"),
)


def get_environment() -> Dict[str, str]:
    """Return a safe to use copy of the system's environment.
    Values starting with BASH_FUNC can cause issues when written in a text file."""
    return {key: value for key, value in os.environ.items() if not key.startswith("BASH_FUNC")}


def execute(
    command: List[str],
    env: Optional[Dict[str, str]] = None,
    cwd: Optional[str] = None,
    quiet: bool = False,
    shell: bool = False,
    timeout: Optional[float] = None,
) -> str:
    """
    Execute a system command and return its standard output; standard error is discarded.

    Params:
        command (list): A list containing an executable and its parameters
        env (dict): Dict of values to add to the current environment
        cwd (str): Working directory
        quiet (bool): Do not display log messages
        timeout (int): Number of seconds the program is allowed to run, disabled by default

    Returns:
        str: stdout output
    """
    stdout, _stderr = _execute(command, env=env, cwd=cwd, quiet=quiet, shell=shell, timeout=timeout)
    return stdout


def execute_with_error(
    command: List[str],
    env: Optional[Dict[str, str]] = None,
    cwd: Optional[str] = None,
    quiet: bool = False,
    shell: bool = False,
    timeout: Optional[float] = None,
) -> Tuple[str, str]:
    """
    Execute a system command and return its standard output and; standard error in a tuple.

    Params:
        command (list): A list containing an executable and its parameters
        env (dict): Dict of values to add to the current environment
        cwd (str): Working directory
        quiet (bool): Do not display log messages
        timeout (int): Number of seconds the program is allowed to run, disabled by default

    Returns:
        str, str: stdout output and stderr output
    """
    return _execute(command, env=env, cwd=cwd, capture_stderr=True, quiet=quiet, shell=shell, timeout=timeout)


def _execute(
    command: List[str],
    env: Optional[Dict[str, str]] = None,
    cwd: Optional[str] = None,
    capture_stderr: bool = False,
    quiet: bool = False,
    shell: bool = False,
    timeout: Optional[float] = None,
) -> Tuple[str, str]:
    # Check if the executable exists
    if not command:
        logger.error("No executable provided!")
        return "", ""
    if os.path.isabs(command[0]) and not path_exists(command[0]):
        logger.error("No executable found in %s", command)
        return "", ""

    if not quiet:
        logger.debug("Executing %s", " ".join([str(i) for i in command]))

    # Set up environment
    existing_env = get_environment()
    if env:
        if not quiet:
            logger.debug(" ".join("{}={}".format(k, v) for k, v in env.items()))
        env = {k: v for k, v in env.items() if v is not None}
        existing_env.update(env)

    # Piping stderr can cause slowness in the programs, use carefully
    # (especially when using regedit with wine)
    try:
        with subprocess.Popen(
            command,
            shell=shell,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE if capture_stderr else subprocess.DEVNULL,
            env=existing_env,
            cwd=cwd,
            errors="replace",
        ) as command_process:
            stdout, stderr = command_process.communicate(timeout=timeout)
    except (OSError, TypeError) as ex:
        logger.error("Could not run command %s (env: %s): %s", command, env, ex)
        return "", ""
    except subprocess.TimeoutExpired:
        logger.error("Command %s after %s seconds", command, timeout)
        return "", ""

    return stdout.strip(), (stderr or "").strip()


def spawn(
    command: Sequence[str], env: Dict[str, str] = None, cwd: str = None, quiet: bool = False, shell: bool = False
) -> None:
    """
    Execute a system command but discard its results and do not wait
    for it to complete.

    Params:
        command (list): A list containing an executable and its parameters
        env (dict): Dict of values to add to the current environment
        cwd (str): Working directory
        quiet (bool): Do not display log messages
    """

    # Check if the executable exists
    if not command:
        logger.error("No executable provided!")
        return
    if os.path.isabs(command[0]) and not path_exists(command[0]):
        logger.error("No executable found in %s", command)
        return

    if not quiet:
        logger.debug("Spawning %s", " ".join([str(i) for i in command]))

    # Set up environment
    existing_env = get_environment()
    if env:
        if not quiet:
            logger.debug(" ".join("{}={}".format(k, v) for k, v in env.items()))
        env = {k: v for k, v in env.items() if v is not None}
        existing_env.update(env)

    # Piping stderr can cause slowness in the programs, use carefully
    # (especially when using regedit with wine)
    try:
        subprocess.Popen(  # pylint: disable=consider-using-with
            command, shell=shell, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, env=existing_env, cwd=cwd
        )
    except (OSError, TypeError) as ex:
        logger.error("Could not run command %s (env: %s): %s", command, env, ex)


def read_process_output(
    command: Sequence[str], timeout: int = 5, env: Mapping[str, str] = None, error_result: Optional[str] = ""
) -> str:
    """Return the output of a command as a string; if 'error_result' is not None,
    returns that on errors. If it is, raises an exception instead."""
    try:
        return subprocess.check_output(command, timeout=timeout, env=env, encoding="utf-8", errors="ignore").strip()
    except (OSError, subprocess.CalledProcessError, subprocess.TimeoutExpired) as ex:
        logger.error("%s command failed: %s", command, ex)
        if error_result is None:
            raise
        return error_result


def get_md5_in_zip(filename: str) -> str:
    """Return the md5 hash of a file in a zip"""
    with zipfile.ZipFile(filename, "r") as archive:
        files = archive.namelist()
        if len(files) > 1:
            logger.warning("More than 1 file in archive %s, reading 1st one: %s", filename, files[0])
        with archive.open(files[0]) as file_in_zip:
            _hash = read_file_md5(file_in_zip)
    return _hash


def get_md5_hash(filename: str) -> Union[bool, str]:
    """Return the md5 hash of a file."""
    try:
        with open(filename, "rb") as _file:
            _hash = read_file_md5(_file)
    except IOError:
        logger.warning("Error reading %s", filename)
        return False
    return _hash


def read_file_md5(filedesc: IO[bytes]) -> str:
    md5 = hashlib.md5()
    for chunk in iter(lambda: filedesc.read(8192), b""):
        md5.update(chunk)
    return md5.hexdigest()


def get_file_checksum(filename: str, hash_type: str) -> str:
    """Return the checksum of type `hash_type` for a given filename"""
    hasher = hashlib.new(hash_type)
    with open(filename, "rb") as input_file:
        for chunk in iter(lambda: input_file.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


def is_executable(exec_path: str) -> bool:
    """Return whether exec_path is an executable"""
    return os.access(exec_path, os.X_OK)


def make_executable(exec_path: str) -> None:
    file_stats = os.stat(exec_path)
    os.chmod(exec_path, file_stats.st_mode | stat.S_IEXEC)


def can_find_executable(exec_name: str) -> bool:
    """Checks if an executable can be located; if false,
    find_executable will return None, and find_required_executable
    will raise an exception."""
    return bool(find_executable(exec_name))


def find_executable(exec_name: str) -> Optional[str]:
    """Return the absolute path of an executable, or None if
    it could not be found."""
    return shutil.which(exec_name) if exec_name else None


def find_required_executable(exec_name: str) -> str:
    """Return the absolute path of an executable, but raises a
    MissingExecutableError if it could not be found."""
    exe = find_executable(exec_name)
    if not exe:
        raise MissingExecutableError(_("The executable '%s' could not be found.") % exec_name)
    return exe


def get_pid(program: str, multiple: bool = False) -> Optional[Union[str, List[str]]]:
    """Return pid of process.

    :param str program: Name of the process.
    :param bool multiple: If True and multiple instances of the program exist,
        return all of them; if False only return the first one.
    """
    pids = execute(["pgrep", program])
    if not pids.strip():
        return None
    pids = pids.split()
    if multiple:
        return pids
    return pids[0]


def python_identifier(unsafe_string: str) -> str:
    """Converts a string to something that can be used as a python variable"""

    def _dashrepl(matchobj: re.Match[str]) -> str:
        return matchobj.group(0).replace("-", "_")

    return re.sub(r"(\${)([\w-]*)(})", _dashrepl, unsafe_string)


def substitute(string_template: str, variables: Dict[str, str]) -> str:
    """Expand variables on a string template

    Args:
        string_template (str): template with variables preceded by $
        variables (dict): mapping of variable identifier > value

    Return:
        str: String with substituted values
    """
    string_template = python_identifier(str(string_template))
    identifiers = variables.keys()

    # We support dashes in identifiers but they are not valid in python
    # identifiers, which is a requirement for the templating engine we use
    # Replace the dashes with underscores in the mapping and template
    variables = dict((k.replace("-", "_"), v) for k, v in variables.items())
    for identifier in identifiers:
        string_template = string_template.replace("${}".format(identifier), "${}".format(identifier.replace("-", "_")))

    template = string.Template(string_template)
    if string_template in list(variables.keys()):
        return variables[string_template]
    return template.safe_substitute(variables)


def merge_folders(source: str, destination: str) -> None:
    """Merges the content of source to destination"""
    logger.debug("Merging %s into %s", source, destination)
    # We do not use shutil.copytree() here because that would copy
    # the file permissions, and we do not want them.
    source = os.path.abspath(source)
    for dirpath, dirnames, filenames in os.walk(source):
        source_relpath = dirpath[len(source) :].strip("/")
        dst_abspath = os.path.join(destination, source_relpath)
        for dirname in dirnames:
            new_dir = os.path.join(dst_abspath, dirname)
            logger.debug("creating dir: %s", new_dir)
            try:
                os.mkdir(new_dir)
            except OSError:
                pass
        for filename in filenames:
            # logger.debug("Copying %s", filename)
            if not os.path.exists(dst_abspath):
                os.makedirs(dst_abspath)
            shutil.copy(os.path.join(dirpath, filename), os.path.join(dst_abspath, filename), follow_symlinks=False)


def remove_folder(
    path: str,
    completion_function: Optional[TrashPortal.CompletionFunction] = None,
    error_function: Optional[TrashPortal.ErrorFunction] = None,
) -> None:
    """Trashes a folder specified by path, asynchronously. The folder
    likely exists after this returns, since it's using DBus to ask
    for the entrashification.
    """
    if not os.path.exists(path):
        logger.warning("Non existent path: %s", path)
        if completion_function:
            completion_function()
        return
    if os.path.samefile(os.path.expanduser("~"), path):
        if error_function:
            error_function(RuntimeError("Lutris tried to trash home directory!"))
        return

    logger.debug("Trashing folder %s", path)
    TrashPortal([path], completion_function=completion_function, error_function=error_function)


def delete_folder(path: str) -> bool:
    """Delete a folder specified by path immediately. The folder will not
    be recoverable, so consider remove_folder() instead.

    Returns true if the folder was successfully deleted.
    """
    if not os.path.exists(path):
        logger.warning("Non existent path: %s", path)
        return False
    if os.path.samefile(os.path.expanduser("~"), path):
        raise RuntimeError("Lutris tried to erase home directory!")
    logger.debug("Deleting folder %s", path)
    try:
        shutil.rmtree(path)
    except OSError as ex:
        logger.error("Failed to delete folder %s: %s (Error code %s)", path, ex.strerror, ex.errno)
        return False
    return True


def create_folder(path: str) -> Optional[str]:
    """Creates a folder specified by path"""
    if not path:
        return None
    path = os.path.expanduser(path)
    os.makedirs(path, exist_ok=True)
    return path


def list_unique_folders(folders: Iterable[str]) -> Iterable[str]:
    """Deduplicate directories with the same Device.Inode"""
    unique_dirs = {}
    for folder in folders:
        folder_stat = os.stat(folder)
        identifier = "%s.%s" % (folder_stat.st_dev, folder_stat.st_ino)
        if identifier not in unique_dirs:
            unique_dirs[identifier] = folder
    return unique_dirs.values()


def is_removeable(path: str, system_config: Dict[str, str]) -> bool:
    """Check if a folder is safe to remove (not system or home, ...). This needs the
    system config dict so it can check the default game path, too."""
    if not path_exists(path):
        return False

    parts = path.strip("/").split("/")

    if not parts:
        return False

    if parts[0] == "var":
        # Fedora Silverblue puts mount points under /var since they are mutable
        # so we'll special case /var/mnt/<drive>/*.
        if len(parts) > 3 and parts[1] in ("mnt", "media"):
            return True
        return False

    if parts[0] in ("usr", "lib", "etc", "boot", "sbin", "bin"):
        # Path is part of the system folders
        return False

    if parts[0] == "home":
        if len(parts) <= 2:
            return False
        if len(parts) == 3 and parts[2] in PROTECTED_HOME_FOLDERS:
            return False

    if system_config:
        default_game_path = system_config.get("game_path")
        if path_contains(path, default_game_path, resolve_symlinks=False):
            return False

    return True


def fix_path_case(path: str) -> str:
    """Do a case-insensitive check, return the real path with correct case. If the path is
    not for a real file, this corrects as many components as do exist."""
    if not path or os.path.exists(path) or not path.startswith("/"):
        # If a path isn't provided, or it exists as is, or is a relative path, just return it.
        return path
    parts = path.strip("/").split("/")
    current_path = "/"
    for part in parts:
        parent_path = current_path
        current_path = os.path.join(current_path, part)
        if not os.path.exists(current_path) and os.path.isdir(parent_path):
            try:
                path_contents = os.listdir(parent_path)
            except OSError:
                logger.error("Can't read contents of %s", parent_path)
                path_contents = []
            for filename in path_contents:
                if filename.lower() == part.lower():
                    current_path = os.path.join(parent_path, filename)
                    break

    # Only return the path if we got the same number of elements
    if len(parts) == len(current_path.strip("/").split("/")):
        return current_path
    # otherwise return original path
    return path


def get_pids_using_file(path: str) -> Set[str]:
    """Return a set of pids using file `path`."""
    if not os.path.exists(path):
        logger.error("Can't return PIDs using non existing file: %s", path)
        return set()

    fuser_path = find_executable("fuser")
    if not fuser_path:
        logger.warning("fuser not available, please install psmisc")
        return set()
    fuser_output = execute([fuser_path, path], quiet=True)
    return set(fuser_output.split())


def reverse_expanduser(path: str) -> str:
    """Replace '/home/username' with '~' in given path."""
    if not path:
        return path
    user_path = os.path.expanduser("~")
    if path.startswith(user_path):
        path = path[len(user_path) :].strip("/")
        return "~/" + path
    return path


def path_contains(parent: Optional[str], child: Optional[str], resolve_symlinks: bool = False) -> bool:
    """Tests if a child path is actually within a parent directory
    or a subdirectory of it. Resolves relative paths, and ~, and
    optionally symlinks."""

    if parent is None or child is None:
        return False

    resolved_parent = Path(os.path.abspath(os.path.expanduser(parent)))
    resolved_child = Path(os.path.abspath(os.path.expanduser(child)))

    if resolve_symlinks:
        resolved_parent = resolved_parent.resolve()
        resolved_child = resolved_child.resolve()

    return resolved_child == resolved_parent or resolved_parent in resolved_child.parents


def path_exists(path: str, check_symlinks: bool = False, exclude_empty: bool = False) -> bool:
    """Wrapper around system.path_exists that doesn't crash with empty values

    Params:
        path (str): File to the file to check
        check_symlinks (bool): If the path is a broken symlink, return False
        exclude_empty (bool): If true, consider 0 bytes files as non existing
    """
    if not path:
        return False
    if path.startswith("~"):
        path = os.path.expanduser(path)
    if os.path.exists(path):
        if exclude_empty:
            return os.stat(path).st_size > 0
        return True
    if os.path.islink(path):
        logger.warning("%s is a broken link", path)
        return not check_symlinks
    return False


def create_symlink(source: str, destination: str) -> None:
    """Create a symlink from source to destination.
    If there is already a symlink at the destination and it is broken, it will be deleted."""
    is_directory = os.path.isdir(source)
    if os.path.islink(destination) and not os.path.exists(destination):
        logger.warning("Deleting broken symlink %s", destination)
        os.remove(destination)
    try:
        os.symlink(source, destination, target_is_directory=is_directory)
    except OSError:
        logger.error("Failed linking %s to %s", source, destination)


def reset_library_preloads() -> None:
    """Remove library preloads from environment"""
    for key in ("LD_LIBRARY_PATH", "LD_PRELOAD"):
        if os.environ.get(key):
            try:
                del os.environ[key]
            except OSError:
                logger.error("Failed to delete environment variable %s", key)


def get_existing_parent(path: str) -> Optional[str]:
    """Return the 1st existing parent for a folder (or itself if the path
    exists and is a directory). returns None, when none of the parents exists.
    """
    if path == "":
        return None
    if os.path.exists(path) and not os.path.isfile(path):
        return path
    return get_existing_parent(os.path.dirname(path))


def update_desktop_icons() -> None:
    """Update Icon for GTK+ desktop manager
    Other desktop manager icon cache commands must be added here if needed
    """
    if can_find_executable("gtk-update-icon-cache"):
        execute(["gtk-update-icon-cache", "-tf", os.path.join(GLib.get_user_data_dir(), "icons/hicolor")], quiet=True)
        execute(["gtk-update-icon-cache", "-tf", os.path.join(settings.RUNTIME_DIR, "icons/hicolor")], quiet=True)


def get_disk_size(path: str) -> int:
    """Return the disk size in bytes of a file or folder"""

    def get_file_size(file_path: str) -> int:
        return os.stat(file_path).st_size

    if os.path.isfile(path):
        return get_file_size(path)

    total_size = 0
    for base, _dirs, files in os.walk(path):
        paths = [os.path.join(base, f) for f in files]
        total_size += sum(get_file_size(p) for p in paths if os.path.isfile(p) and not os.path.islink(p))
    return total_size


def get_locale_list() -> List[str]:
    """Return list of available locales"""
    try:
        with subprocess.Popen(["locale", "-a"], stdout=subprocess.PIPE) as locale_getter:
            output = locale_getter.communicate()
        locales = output[0].decode("ASCII").split()  # locale names use only ascii characters
    except FileNotFoundError:
        lang = os.environ.get("LANG", "")
        if lang:
            locales = [lang]
        else:
            locales = []
    return locales


def get_running_pid_list() -> List[int]:
    """Return the list of PIDs from processes currently running"""
    return [int(p) for p in os.listdir("/proc") if p[0].isdigit()]


def get_mounted_discs() -> List[Optional[str]]:
    """Return a list of mounted discs and ISOs

    :rtype: list of Gio.Mount
    """
    volumes = Gio.VolumeMonitor.get()
    drives = []

    for mount in volumes.get_mounts():
        volume = mount.get_volume()
        if volume:
            device = volume.get_identifier("unix-device")
            if not device:
                logger.debug("No device for mount %s", mount.get_name())
                continue

            # Device is a disk drive or ISO image
            if "/dev/sr" in device or "/dev/loop" in device:
                drives.append(mount.get_root().get_path())
    return drives


def find_mount_point(path: str) -> str:
    """Return the mount point a file is located on"""
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path


def set_keyboard_layout(layout: str) -> None:
    setxkbmap_command = ["setxkbmap", "-model", "pc101", layout, "-print"]
    xkbcomp_command = ["xkbcomp", "-", os.environ.get("DISPLAY", ":0")]
    with subprocess.Popen(xkbcomp_command, stdin=subprocess.PIPE) as xkbcomp:
        with subprocess.Popen(setxkbmap_command, env=os.environ, stdout=xkbcomp.stdin) as setxkbmap:
            setxkbmap.communicate()
            xkbcomp.communicate()



import os
import re
import shutil
import subprocess
import tempfile
import xml.etree.ElementTree

from lutris.util.log import logger
from lutris.util.system import execute, read_process_output


class CabInstaller:
    """Extract and install contents of cab files

    Based on an implementation by tonix64: https://github.com/tonix64/python-installcab
    """

    def __init__(self, prefix, arch=None, wine_path=None):
        self.prefix = prefix
        self.winearch = arch or self.get_wineprefix_arch()
        self.tmpdir = tempfile.mkdtemp()
        self.wine_path = wine_path

        self.register_dlls = False  # Whether to register DLLs, I don't the purpose of that
        self.strip_dlls = False  # When registering, strip the full path

    @staticmethod
    def process_key(key):
        """I have no clue why"""
        return key.strip("\\").replace("HKEY_CLASSES_ROOT", "HKEY_LOCAL_MACHINE\\Software\\Classes")

    @staticmethod
    def get_arch_from_manifest(root):
        registry_keys = root.findall("{urn:schemas-microsoft-com:asm.v3}assemblyIdentity")
        arch = registry_keys[0].attrib["processorArchitecture"]
        arch_map = {"amd64": "win64", "x86": "win32", "wow64": "wow64"}
        return arch_map[arch]

    def get_winebin(self, arch):
        wine_path = self.wine_path or "wine"
        return wine_path if arch in ("win32", "wow64") else wine_path + "64"

    @staticmethod
    def get_arch_from_dll(dll_path):
        if "x86-64" in read_process_output(["file", dll_path]):
            return "win64"
        return "win32"

    def cleanup(self):
        logger.info("Cleaning up %s", self.tmpdir)
        shutil.rmtree(self.tmpdir)

    def check_dll_arch(self, dll_path):
        return self.get_arch_from_dll(dll_path)

    def replace_variables(self, value, arch):
        if "$(" in value:
            value = value.replace("$(runtime.help)", "C:\\windows\\help")
            value = value.replace("$(runtime.inf)", "C:\\windows\\inf")
            value = value.replace("$(runtime.wbem)", "C:\\windows\\wbem")
            value = value.replace("$(runtime.windows)", "C:\\windows")
            value = value.replace("$(runtime.ProgramFiles)", "C:\\windows\\Program Files")
            value = value.replace("$(runtime.programFiles)", "C:\\windows\\Program Files")
            value = value.replace("$(runtime.programFilesX86)", "C:\\windows\\Program Files (x86)")
            value = value.replace("$(runtime.system32)", "C:\\windows\\%s" % self.get_system32_realdir(arch))
            value = value.replace(
                "$(runtime.drivers)",
                "C:\\windows\\%s\\drivers" % self.get_system32_realdir(arch),
            )
        value = value.replace("\\", "\\\\")
        return value

    def process_value(self, reg_value, arch):
        attrs = reg_value.attrib
        name = attrs["name"]
        value = attrs["value"]
        value_type = attrs["valueType"]
        if not name.strip():
            name = "@"
        else:
            name = '"%s"' % name
        name = self.replace_variables(name, arch)
        if value_type == "REG_BINARY":
            value = re.findall("..", value)
            value = "hex:" + ",".join(value)
        elif value_type == "REG_DWORD":
            value = "dword:%s" % value.replace("0x", "")
        elif value_type == "REG_QWORD":
            value = "qword:%s" % value.replace("0x", "")
        elif value_type == "REG_NONE":
            value = None
        elif value_type == "REG_EXPAND_SZ":
            # not sure if we should replace this ones at this point:
            # caps can vary in the pattern
            value = value.replace("%SystemRoot%", "C:\\windows")
            value = value.replace("%ProgramFiles%", "C:\\windows\\Program Files")
            value = value.replace("%WinDir%", "C:\\windows")
            value = value.replace("%ResourceDir%", "C:\\windows")
            value = value.replace("%Public%", "C:\\users\\Public")
            value = value.replace("%LocalAppData%", "C:\\windows\\Public\\Local Settings\\Application Data")
            value = value.replace("%AllUsersProfile%", "C:\\windows")
            value = value.replace("%UserProfile%", "C:\\windows")
            value = value.replace("%ProgramData%", "C:\\ProgramData")
            value = '"%s"' % value
        elif value_type == "REG_SZ":
            value = '"%s"' % value
        else:
            logger.warning("warning unkown type: %s", value_type)
            value = '"%s"' % value
        if value:
            value = self.replace_variables(value, arch)
            if self.strip_dlls:
                if ".dll" in value:
                    value = value.lower().replace("c:\\\\windows\\\\system32\\\\", "")
                    value = value.lower().replace("c:\\\\windows\\\\syswow64\\\\", "")
        return name, value

    def get_registry_from_manifest(self, file_name):
        out = ""
        root = xml.etree.ElementTree.parse(file_name).getroot()
        arch = self.get_arch_from_manifest(root)
        registry_keys = root.findall("{urn:schemas-microsoft-com:asm.v3}registryKeys")
        if registry_keys:
            for registry_key in list(registry_keys[0]):
                key = self.process_key(registry_key.attrib["keyName"])
                out += "[%s]\n" % key
                for reg_value in registry_key.findall("{urn:schemas-microsoft-com:asm.v3}registryValue"):
                    name, value = self.process_value(reg_value, arch)
                    if value is not None:
                        out += "%s=%s\n" % (name, value)
                out += "\n"
        return (out, arch)

    def get_wineprefix_arch(self):
        with open(os.path.join(self.prefix, "system.reg"), encoding="utf-8") as reg_file:
            for line in reg_file.readlines():
                if line.startswith("#arch=win32"):
                    return "win32"
                if line.startswith("#arch=win64"):
                    return "win64"
        return "win64"

    def get_system32_realdir(self, arch):
        dest_map = {
            ("win64", "win32"): "Syswow64",
            ("win64", "win64"): "System32",
            ("win64", "wow64"): "System32",
            ("win32", "win32"): "System32",
        }
        return dest_map[(self.winearch, arch)]

    def get_dll_destdir(self, dll_path):
        if self.get_arch_from_dll(dll_path) == "win32" and self.winearch == "win64":
            return os.path.join(self.prefix, "drive_c/windows/syswow64")
        return os.path.join(self.prefix, "drive_c/windows/system32")

    def install_dll(self, dll_path):
        dest_dir = self.get_dll_destdir(dll_path)
        logger.debug("Copying %s to %s", dll_path, dest_dir)
        shutil.copy(dll_path, dest_dir)

        dest_dll_path = os.path.join(dest_dir, os.path.basename(dll_path))
        if not self.register_dlls:
            return
        arch = self.get_arch_from_dll(dest_dll_path)
        subprocess.call([self.get_winebin(arch), "regsvr32", os.path.basename(dest_dll_path)])

    def get_registry_files(self, output_files):
        reg_files = []
        for file_path in output_files:
            if file_path.endswith(".manifest"):
                out = "Windows Registry Editor Version 5.00\n\n"
                outdata, arch = self.get_registry_from_manifest(file_path)
                if outdata:
                    out += outdata
                    with open(os.path.join(self.tmpdir, file_path + ".reg"), "w", encoding="utf-8") as reg_file:
                        reg_file.write(out)
                    reg_files.append((file_path + ".reg", arch))
            if file_path.endswith(".dll"):
                self.install_dll(file_path)
        return reg_files

    def apply_to_registry(self, file_path, arch):
        logger.info("Applying %s to registry", file_path)
        subprocess.call([self.get_winebin(arch), "regedit", os.path.join(self.tmpdir, file_path)])

    def extract_from_cab(self, cabfile, component):
        """Extracts files matching a `component` name from a `cabfile`

        Params:
            cabfile (str): Path to a cabfile to extract from
            component (str): component to extract from the cab file

        Returns:
            list: Files extracted from the cab file
        """
        execute(["cabextract", "-F", "*%s*" % component, "-d", self.tmpdir, cabfile])
        return [os.path.join(r, file) for r, d, f in os.walk(self.tmpdir) for file in f]

    def install(self, cabfile, component):
        """Install `component` from `cabfile`"""
        logger.info("Installing %s from %s", component, cabfile)

        for file_path, arch in self.get_registry_files(self.extract_from_cab(cabfile, component)):
            self.apply_to_registry(file_path, arch)

        self.cleanup()



from gi.repository import Gtk, Pango

from lutris.util.strings import is_valid_pango_markup


class InstallerLabel(Gtk.Label):
    """A label for installers"""

    def __init__(self, text, wrap=True, selectable=False):
        super().__init__()

        if wrap:
            self.set_line_wrap(True)
            self.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        else:
            self.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        self.set_alignment(0, 0.5)
        self.set_margin_right(12)

        if is_valid_pango_markup(text):
            self.set_markup(text)
        else:
            self.set_text(text)

        self.props.can_focus = False
        self.set_tooltip_text(text)
        self.set_selectable(selectable)
